import prompt_toolkit.application as ptk_app
import prompt_toolkit.formatted_text as ptk_formatted_text
import prompt_toolkit.layout.containers as ptk_containers
import prompt_toolkit.layout.controls as ptk_controls
import prompt_toolkit.layout.layout as ptk_layout

from pygments.token import Token as T

from .multi import MultiForm


class CheckboxForm(MultiForm):
    """Form for when the user can select multiple options from Choices.

    This form is different from ListForm in that the user can select
    multiple options. As such, the answer is stored as a list().
    """

    def __init__(self, **kwargs: dict) -> None:
        super(CheckboxForm, self).__init__(**kwargs)
        self._selected = set()
        self._add_key_bindings()

    def _add_key_bindings(self) -> None:
        """Adds keys to self.__key_bindings for list movements."""

        @self._key_bindings.add("enter")
        def enter(event: object) -> None:
            self._selection = [self._values[s] for s in sorted(self._selected)]
            validation = (
                self._validate(self._selection) if self._validate else True
            )
            if validation is True:
                event.app.exit()
            else:
                message = "Could not validate the input succesfully."
                if isinstance(validation, str):
                    message = validation
                self._display_error(message)

        @self._key_bindings.add("space")
        def click(event: object) -> None:
            if self._idx_cursor not in self._selected:
                self._selected.add(self._idx_cursor)
                self._select(self._choices_windows[self._idx_cursor])
            else:
                self._selected.remove(self._idx_cursor)
                self._deselect(self._choices_windows[self._idx_cursor])

        @self._key_bindings.add("a")
        def select_all(event: object) -> None:
            for idx in range(len(self._choices_windows)):
                self._selected.add(idx)
                self._select(self._choices_windows[idx])

        @self._key_bindings.add("i")
        def invert_all(event: object) -> None:
            for idx in range(len(self._choices_windows)):
                if idx not in self._selected:
                    self._selected.add(idx)
                    self._select(self._choices_windows[idx])
                else:
                    self._selected.remove(idx)
                    self._deselect(self._choices_windows[idx])

    def _select(self, window: ptk_containers.Window) -> None:
        """Marks a given choice as selected."""
        token = (T.Selector, "● ")
        window.content.text.token_list[1] = token

    def _deselect(self, window: ptk_containers.Window) -> None:
        """Marks a given choice as deselected."""
        token = (T.Selector, "○ ")
        window.content.text.token_list[1] = token

    def _display_error(self, message: str) -> None:
        """Displays an error message in a dedicated Window.

        This is used when the validation fails to notify the user.

        Args:
            message: The message to display the user.
        """
        token = (T.Error, message)
        self._error_window.content.text.token_list[0] = token
        self._error_window.height = 1

    def _generate_error_window(self):
        """Generates an empty error Window for later use."""
        #  Note that because the height is 0 by default, this window
        #  will not be visible unless the height is changed.
        empty_tokens = [(T.Error, "")]
        empty_fragments = ptk_formatted_text.PygmentsTokens(empty_tokens)
        empty_text = ptk_controls.FormattedTextControl(
            empty_fragments, show_cursor=False
        )
        return ptk_containers.Window(empty_text, height=0)

    def _ask_question(self, answers: dict) -> None:
        """Asks a question and store the answer in the answers dict."""
        instructions = (
            "(<up>, <down> to move, <space> to select, "
            "<a> to select all, <i> to invert all)"
        )
        self._question_window = self._generate_question_window(instructions)
        self._choices_windows = self._generate_choices_windows("Checkbox")
        self._error_window = self._generate_error_window()
        windows = (
            [self._question_window]
            + self._choices_windows
            + [self._error_window]
        )
        body = ptk_containers.HSplit(windows)
        application = ptk_app.Application(
            layout=ptk_layout.Layout(body),
            key_bindings=self._key_bindings,
            full_screen=False,
            style=self._style,
        )
        application.run()
        answers[self._name] = self._selection



import prompt_toolkit as ptk
import prompt_toolkit.formatted_text as ptk_formatted_text
import prompt_toolkit.keys as ptk_keys
from pygments.token import Token as T

from .input import InputForm


class ConfirmForm(InputForm):
    """A form where the user is asked to confirm with a Yes/No question.

    This form is a more specific version of InputForm, one where pressing
    one between Y, y, N and n will immediately stop the prompt and
    store either a True or False into answers.
    """

    def __init__(self, **kwargs):
        super(ConfirmForm, self).__init__(**kwargs)
        self._add_key_bindings()

    def _add_key_bindings(self) -> None:
        """Adds keys to self.__key_bindings for y/N confirmation."""

        @self._key_bindings.add("y")
        @self._key_bindings.add("Y")
        def yes(event) -> None:
            event.current_buffer.text = "y"
            event.app.exit(result=True)

        @self._key_bindings.add("n")
        @self._key_bindings.add("N")
        def no(event) -> None:
            event.current_buffer.text = "n"
            event.app.exit(result=False)

        @self._key_bindings.add(ptk_keys.Keys.Any)
        def _(event) -> None:
            """Disallows inserting other text."""
            pass

    def _format_message(self) -> ptk_formatted_text.PygmentsTokens:
        message_fragments = super(ConfirmForm, self)._format_message()
        instructions_tokens = [(T.Instruction, "(y/n) ")]
        message_fragments.token_list += instructions_tokens
        return message_fragments

    def _ask_question(self, answers: dict) -> None:
        answers[self._name] = ptk.prompt(
            self._format_message(),
            style=self._style,
            validate_while_typing=False,
            key_bindings=self._key_bindings,
        )



import typing as t

import prompt_toolkit as ptk
import prompt_toolkit.document as ptk_document
import prompt_toolkit.formatted_text as ptk_formatted_text
import prompt_toolkit.validation as ptk_validation
from pygments.token import Token as T

from .abstract import AbstractForm


class InputValidator(ptk_validation.Validator):
    """Validator wrapper for InputForms.

    The prompt() in InputForms accept a ptk_validation.Validator as
    argument for validation (which means we don't need to rely on
    custom-made solutions). Thus, the function provided by the user
    is wrapped in a InputValidator. The usual applies: if the
    function returns True, the input is validated. If False or str,
    the validation fails (if str, then the string is used as the
    message to display to the user).

    Args:
        function: The function to use for validation.
    """

    def __init__(self, function: t.Callable[[str], t.Union[bool, str]]):
        self._function = function

    def validate(self, document: ptk_document.Document) -> None:
        """Validates the content inputted by the user."""
        validation = self._function(document.text)
        cursor_position = len(document.text)

        message = ""
        if not validation:
            message = "The input was not validated succesfully."
        if isinstance(validation, str):
            message = validation
        if message:
            raise ptk_validation.ValidationError(
                message=message, cursor_position=cursor_position
            )


class InputForm(AbstractForm):
    """Class for forms of type Input.

    InputForms are forms where the user is asked to type some input
    following a question, e.g. [?] What's your name?
    """

    def __init__(self, **kwargs: dict) -> None:
        super(InputForm, self).__init__(**kwargs)
        if self._validate:
            self._validator = InputValidator(self._validate)
        else:
            self._validator = None

    def _format_message(self) -> ptk_formatted_text.PygmentsTokens:
        """Formats the message provided by the user to improve readability."""
        message_tokens = [
            (T.QuestionMark, "[?] "),
            (T.Question, self._message + " "),
        ]
        message_fragments = ptk_formatted_text.PygmentsTokens(message_tokens)
        return message_fragments

    def _ask_question(self, answers: dict) -> None:
        answers[self._name] = ptk.prompt(
            self._format_message(),
            style=self._style,
            validator=self._validator,
            validate_while_typing=False,
        )



import typing as t

import prompt_toolkit.application as ptk_app
import prompt_toolkit.formatted_text as ptk_formatted_text
import prompt_toolkit.layout.containers as ptk_containers
import prompt_toolkit.layout.controls as ptk_controls
from pygments.token import Token as T

from .abstract import AbstractForm


class MultiForm(AbstractForm):
    """Parent class for forms that presents multiple options to choose from.

    Both ListForm and CheckboxForm are children of this class. This parent
    class implements a few comodity functions such as vertical movements
    for the cursor.
    """

    #  Parent class for Checkbox and Listform.
    def __init__(self, **kwargs: dict) -> None:
        super(MultiForm, self).__init__(**kwargs)
        self._idx_cursor = 0
        self._add_cursor_key_bindings()

    def _add_cursor_key_bindings(self) -> None:
        """Adds keys to self.__key_bindings for up/down movements."""

        @self._key_bindings.add("up")
        def move_cursor_up(event: object) -> None:
            self._move_cursor_up(event.app)

        @self._key_bindings.add("down")
        def move_cursor_down(event: object) -> None:
            self._move_cursor_down(event.app)

    def _unset_cursor(self, window: ptk_containers.Window) -> None:
        """Removes the cursor pointer from a given window."""
        token = (T.Pointer, "  ")
        window.content.text.token_list[0] = token

    def _set_cursor(self, window: ptk_containers.Window) -> None:
        """Adds the cursor pointer to a given window."""
        token = (T.Pointer, "❯ ")
        window.content.text.token_list[0] = token

    def _move_cursor_up(self, application: ptk_app.Application) -> None:
        """Moves the cursor from one window to the one above.

        The cursor is removed from the "active" window (self._idx_cursor
        is used to keep track of which window is active) and added to
        the one above it.
        """
        #  If the cursor is already at first position, then
        #  we don't do anything and simply return None.
        if self._idx_cursor == 0:
            return None
        self._unset_cursor(self._choices_windows[self._idx_cursor])
        self._set_cursor(self._choices_windows[self._idx_cursor - 1])
        self._idx_cursor -= 1

    def _move_cursor_down(self, application: ptk_app.Application) -> None:
        """Moves the cursor from one window to the one below.

        The cursor is removed from the "active" window (self._idx_cursor
        is used to keep track of which window is active) and added to
        the one below it.
        """
        #  If the cursor is already at last position, then
        #  we don't do anything and simply return None.
        max_idx = len(self._choices) - 1
        if self._idx_cursor == max_idx:
            return None
        self._unset_cursor(self._choices_windows[self._idx_cursor])
        self._set_cursor(self._choices_windows[self._idx_cursor + 1])
        self._idx_cursor += 1

    def _generate_question_window(
        self, instructions: str
    ) -> ptk_containers.Window:
        """Generates a Window for the question and instructions."""
        question_tokens = [
            (T.QuestionMark, "[?] "),
            (T.Question, self._message),
            (T.Instruction, " {}".format(instructions)),
        ]
        question_fragments = ptk_formatted_text.PygmentsTokens(question_tokens)
        question_text = ptk_controls.FormattedTextControl(
            question_fragments, show_cursor=False
        )
        return ptk_containers.Window(question_text, height=1)

    def _generate_choices_windows(
        self, form_type: str
    ) -> t.List[ptk_containers.Window]:
        """Generates a list of Windows for the various Choices."""
        assert form_type in ("List", "Checkbox")
        tokens_groups = []
        #  First, each choice is transformed into a Pygments' Token. These
        #  tokens are not supported natively in prompt_toolkit anymore, so
        #  they need to be transformed via PygmentsTokens().
        for idx, choice in enumerate(self._choices):
            tokens = [(T.Pointer, "❯ ")] if idx == 0 else [(T.Pointer, "  ")]
            if form_type == "Checkbox":
                tokens.append((T.Selector, "○ "))
            tokens.append((T.Text, choice))
            tokens_groups.append(tokens)
        fragments_groups = [
            ptk_formatted_text.PygmentsTokens(tokens)
            for tokens in tokens_groups
        ]
        texts = [
            ptk_controls.FormattedTextControl(fragments, show_cursor=False)
            for fragments in fragments_groups
        ]
        return [ptk_containers.Window(text, height=1) for text in texts]



import abc
import collections

import prompt_toolkit.key_binding as ptk_key_binding


class AbstractForm(abc.ABC):
    """Abstract class for all the other Reptile forms.

    Args:
        kwargs: A dictionary containg all the information necessary to
            instantie the form. Specifically:
            - Name: The name of the question being asked. This acts as a
                unique identifier for the question. These names will be
                then used as key for the answers dictionary. This field
                is mandatory.
            - Message: The message to display with the question, it's
                normally the question itself. This field is currently
                mandatory.
            - Choices: An array-like structure of values to display to
                the user. Only relevant for list-like forms such as ListForm.
                Optional for certain forms and required for others.
            - Values: An array-like structure with the same lenght as Choices.
                If specified (it's entirely optional), it provides
                backend values for Choices. If the user select the second
                value in Choices within the form, then the final
                dictionary will store the second element of Values. If
                Values is not specified, then Choices is used instead.
            - Validate: A function to use to validate the data. These
                function can return True (valid), False (invalid with
                generic message) or a str (invalid, the string is used
                as error message). Note that some forms may not use
                Validate even when provided (e.g., ListForm).
            - Transform: A function that takes the input provided
                by the user and transform it into something else. Assuming
                the value provided by the user is x, the final answers dict
                will contain Transfom(x) instead of x.
            - When: A function that takes the whole answers dict (before
                the specific question is asked) and return either True or
                False. If True, the question is asked, if False skipped.
                This is useful if you want to have a series of questions
                with some of them being conditional.
    """

    def __init__(self, **kwargs: dict) -> None:
        k = collections.defaultdict(lambda: None, kwargs)
        self._name = k["Name"]
        self._message = k["Message"]
        self._choices = k["Choices"]
        self._validate = k["Validate"]
        self._transform = k["Transform"]
        self._when = k["When"]
        self._style = k["Style"]
        #  If no Values are passed, self._values default to Choices
        #  if Choices is actually available.
        self._values = k["Values"] if k["Values"] else k["Choices"]
        if self._values:
            assert len(self._values) == len(self._choices)
        if "Default" in k:
            #  If Default is not defined, then self._default is not
            #  specified. The reason we don't assign None to it is that the
            #  user may want to specify None as a default value and we want
            #  to be able to respect that.
            self._default = k["Default"]
        self._generate_key_bindings()

    def _generate_key_bindings(self) -> None:
        """Generates a KeyBindings() object and store it internally.

        Children classes can then directly add keys to the self._key_bindings
        via the apposite decorator @self._key_bindings.add().
        """
        self._key_bindings = ptk_key_binding.KeyBindings()

    @abc.abstractmethod
    def _ask_question(self, answers: dict) -> None:
        """Asks the question and stores the answer in the dict.

        This method is an abstract ABC method, which means it has to be
        overridden by the child class before any object can be instantiated.

        Args:
            answers: the dict where to store the answers.
        """
        pass

    def ask_question(self, answers: dict) -> None:
        """Asks the question and stores the answer in the dict.

        This is the only method that should be public in a form and it
        should never be overriden. On top of asking a question, it
        checks whether the question should be asked (the When), it provides
        a default value if one is specified, etc.

        Args:
            answers: the dict where to store the answers.
        """
        if not self._when or self._when(answers):
            self._ask_question(answers)
            if not answers[self._name] and hasattr(self, "_default"):
                answers[self._name] = self._default
            if self._transform:
                answers[self._name] = self._transform(answers[self._name])



import setuptools

setuptools.setup(
    name="reptile",
    version="1.0.1",
    author="Alessandro",
    url="https://github.com/alessandrosp/reptile",
    description=(
        "Reptile is a command-line interface for Python. Specifically, "
        "Reptile helps with producing interactive REPL-like software."
    ),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords=(
        "prompt-toolkit, cli, command-line, commandline, "
        "command-line-interface, python-inquiry, inquirer, "
        "reptile, REPL, prompt"
    ),
    install_requires=["prompt-toolkit>=3.0.5", "pygments>=2.6.1"],
    python_requires=">=3.6",
    packages=setuptools.find_packages(),
)



import typing as t

import prompt_toolkit.styles as ptk_style
from pygments.token import Token as T

from .forms.checkbox import CheckboxForm
from .forms.confirm import ConfirmForm
from .forms.input import InputForm
from .forms.list import ListForm


#  The default style for the various forms. This can be overwritten
#  by passing a new style to the question (key: 'Style').
DEFAULT_STYLE = ptk_style.style_from_pygments_dict(
    {
        #  Style for the answer once it has been submitted.
        T.Answer: "#FF9D00 bold",
        #  Style for the erorr message (validation).
        T.Error: "#E6E5E6 bg:#5F0000",
        #  Style for the instruction snippets.
        T.Instruction: "",
        #  Style used for the cursor (pointer) in MultiForms.
        T.Pointer: "#FF9D00 bold",
        #  Style used for the text in the question.
        T.Question: "bold",
        #  Style for the text prepended to the question (i.e. [?]).
        T.QuestionMark: "#A4F743 bold",
        # Style for the selector (in CheckboxForm).
        T.Selector: "#FF9D00",
    }
)
#  Only the following types are valid questions' types. They each
#  map to a specific form. If a question is asked with a different
#  type an error is raised.
ACCEPTED_TYPES = ["Checkbox", "Confirm", "Input", "List"]
# Map from the string type to the correspondent class.
FORMS_MAP = {
    "Checkbox": CheckboxForm,
    "Confirm": ConfirmForm,
    "List": ListForm,
    "Input": InputForm,
}


class ReptileError(Exception):
    """Basic error for Reptile."""


class UnnamedQuestion(ReptileError):
    """Raised when one of the question doesn't have the field Name."""


class NotUniqueNames(ReptileError):
    """Raised when two or more questions share the same name."""


class InvalidFormType(ReptileError):
    """Raised when the question's type is not included in ACCEPTED_TYPES."""


class MissingFormType(ReptileError):
    """Raised when one of the questions is missing the field Type."""


def prompt(questions: t.Union[list, dict]) -> dict:
    """The primary function the user should interact with.

    It takes some questions (either as a single dict or a list of dicts),
    creates the relevant froms (depending on the key Type) and store
    the responses in the output dict, answers.

    Args:
        questions: The questions to ask, either as a single dict or
            as a list of dicts.

    Returns:
        The answers dict with contains for each question the relevant
        answer. The answers are under a key named after the Name field
        in the relevant question.
    """

    def _check_questions_are_named(questions: t.List[dict]) -> None:
        """Checks that all questions have a Name field."""
        for question in questions:
            if "Name" not in question or not question["Name"]:
                message = "Every question needs to have a Name."
                raise UnnamedQuestion(message)

    def _check_names_are_unique(questions: t.List[dict]) -> None:
        """Checks that all the names across the list of dicts are unique."""
        list_names = [question["Name"].lower() for question in questions]
        set_names = {question["Name"].lower() for question in questions}
        if len(list_names) != len(set_names):
            message = "The questions' names must be unique."
            raise NotUniqueNames(message)

    def _check_valid_form_types(questions: t.List[dict]) -> None:
        """Checks that the various Type used are in ACCEPTED_TYPES."""
        for question in questions:
            if "Type" not in question:
                message = "Questions must specify the type of form to use."
                raise MissingFormType(message)
            if question["Type"] not in ACCEPTED_TYPES:
                message = "The Type selected is not supported."
                raise InvalidFormType(message)

    if isinstance(questions, dict):
        questions = [questions]
    _check_questions_are_named(questions)
    _check_names_are_unique(questions)
    _check_valid_form_types(questions)
    forms = []
    answers = {}
    for question in questions:
        if "Style" not in question or not question["Style"]:
            question["Style"] = DEFAULT_STYLE
        forms.append(FORMS_MAP[question["Type"]](**question))
    for form in forms:
        form.ask_question(answers)
    return answers



import prompt_toolkit.application as ptk_app
import prompt_toolkit.layout.containers as ptk_containers
import prompt_toolkit.layout.layout as ptk_layout

from .multi import MultiForm


class ListForm(MultiForm):
    """Class for forms of type List.

    ListForms are forms where the user is presented with a series of
    Choices and can only select one of these options. If Values is specified
    then the corresponding value of Values is stored in the answers dict,
    otherwise Choices is used instead.

    Note that ListForms do not implement any validation mechanisms, so
    even if you pass a function nothing will happen. That's because
    the Choices are limited and decided by whoever created the form.
    """

    def __init__(self, **kwargs: dict) -> None:
        super(ListForm, self).__init__(**kwargs)
        self._add_key_bindings()

    def _add_key_bindings(self) -> None:
        """Adds keys to self.__key_bindings for list movements."""

        @self._key_bindings.add("enter")
        def enter(event: object) -> None:
            event.app.exit()

    def _ask_question(self, answers: dict) -> None:
        """Asks a question and store the answer in the answers dict."""
        instructions = "(Use arrow keys)"
        self._question_window = self._generate_question_window(instructions)
        self._choices_windows = self._generate_choices_windows("List")
        windows = [self._question_window] + self._choices_windows
        body = ptk_containers.HSplit(windows)
        application = ptk_app.Application(
            layout=ptk_layout.Layout(body),
            key_bindings=self._key_bindings,
            full_screen=False,
            style=self._style,
        )
        application.run()
        answers[self._name] = self._values[self._idx_cursor]
